"""
ladevorgangclass.py, part of soc_helper
Copyright (C) 2024  M. Williges (spam at zut punkt de)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# Systemimporte
import json
import importlib

# eigene Importe
import log
import datetime
import configuration
if configuration.USE_SPRITMONITOR:
    import spritmonitor

def reloadConfig():
    importlib.invalidate_caches()
    importlib.reload(configuration)

class ladevorgangclass:
    # Definiert Methoden und Daten für einen Ladevorgang.
    def __init__(self):
        self.__savefile = None	        # Lokale Datei zum Speichern der Ladevorgänge
        self.__counter = -1             # Zähler der OpenWB, alle paar Sekunden aktualisiert
        self.__counterAtPlugIn = -1     # OpenWB-Zählerstand beim Einstecken des Ladesteckers
        self.__odo = 0                  # Kilometerstand des Fahrzeugs
        self.__soc = 0                  # der letzte empfangene SOC des Fahrzeugs
        self.__openwbsoc = 0            # der (berechnete) SOC aus der Wallbox
        self.__plug = 0                 # Zustand des Ladesteckers

    def setLogfilehandle(self, savefile):
        self.__savefile = savefile

    def setCounter(self, e):
        self.__counter = e
    
    def setOdo(self, o):
        self.__odo = o
    
    def setSoc(self, s):
        self.__soc = s
    
    def setOpenwbsoc(self, s):
        self.__openwbsoc = float(s)
    
    def setPlug(self, s):
        # A change in plug state may indicate the begin or end of a charge
        if (self.__plug == 0) and (s != 0):
            #Stecker wurde eingesteckt. Energiezähler sichern
            self.__counterAtPlugIn = self.__counter
            log.output(2,'Ladestecker wurde gesteckt bei Zähler '+str(self.__counterAtPlugIn))
        elif (self.__plug !=0) and (s==0):
            #Stecker wurde abgesteckt, merken des Zählerstandes bei Ende der Ladung
            counterAtPlugOut = self.__counter
            log.output(2,'Ladestecker wurde entfernt bei Zähler '+str(counterAtPlugOut))
            # Auswertung des Ladevorgangs.
            datetimeAtPlugOut = datetime.datetime.now()	# Datum des Ladeendes
            # Datum in lesbarer Form für lokale Datei und ggf. Spritmonitor
            date = datetimeAtPlugOut.strftime('%d.%m.%Y')
            # Zähler auf kWh umrechnen und mit 3 Nachkommastellen runden
            geladen = round( (counterAtPlugOut - self.__counterAtPlugIn)/configuration.OPENWB_FACTORCOUNTER, 3)
            log.output(1,'Es wurden '+str(geladen)+'kWh geladen.')
            if self.__savefile != None: 
                # lokales Abspeichern versuchen
                try:
                    self.__savefile.write(date+', '+str(self.__odo)+
                            ', '+str(geladen)+', '+str(self.__soc)+', '+str(self.__openwbsoc)+'\n')
                    self.__savefile.flush()
                except:
                    log.output(0,'lokales Abspeichern fehlgeschlagen!')

            #Spritmonitor-Aufruf, nur wenn konfiguriert und >=100Wh geladen wurden
            if configuration.USE_SPRITMONITOR and (geladen >= 0.1):
                log.output(2,'Spritmonitor ist konfiguriert. Beginne Übermittlung.')
                # letzten bei Spritmonitor eingetragenen km-Stand auslesen
                lastFueling = spritmonitor.get_last_fuel_entry(configuration.SPRITMONITOR_VEHICLE_ID)
                if len(lastFueling) > 0:
                    td = json.loads(json.dumps(lastFueling[0]))
                    lastOdo = float(td['odometer'])
                    log.output(2,'Letzter Spritmonitor-Kilometerstand ist '+str(lastOdo))
                else:
                    lastOdo = 0.0
                    log.output(2,"Konnte keinen letzter Spritmonitor-Kilometerstand ermitteln, nehme 0 an.")
                log.output(2,"Letzter Spritmonitor-Kilometerstand ist "+str(lastOdo))
                log.output(2,"Ladung wurde mit berechneten "+str(self.__openwbsoc)+"% beendet.")
                # type-Variable für Füllung bestimmen
                if lastOdo == 0.0:
                    fuel_type = "first"
                elif ( abs(float(self.__openwbsoc)-100)<=2.0 ):
                    fuel_type = "full" 
                else:
                    fuel_type = "notfull"
                trip = round(self.__odo - lastOdo,2)
                if trip < 0.1:
                    trip = 0.0
                log.output(1,"Letzter Spritmonitor-Kilometerstand: "+str(lastOdo)
                        +'; aktueller Kilometerstand: '+str(self.__odo)+' trip: '+str(trip))
                quantityunitid = 5 # kWh
                # "Betankung" übermitteln
                result=spritmonitor.add_fuel_entry(configuration.SPRITMONITOR_VEHICLE_ID, 1,
                        date, fuel_type, self.__odo, trip, geladen, quantityunitid,
                        configuration.SPRITMONITOR_FUELSORT, configuration.SPRITMONITOR_FUELPRICE,
                        self.__openwbsoc, configuration.SPRITMONITOR_ATTRIBUTES)
                log.output(2,'Ergebnis der Übermittlung: '+str(result))
                log.output(1,'Übermittlung an Spritmonitor ist erfolgt')
        elif (self.__plug != 0) and (self.__counterAtPlugIn<0) and (self.__counter>=0):
            # der Stecker steckte schon, als das Programm gestartet wurde und es liegt noch kein
            # Zählerstand bei Einstecken vor. In diesem Fall wird der aktuelle Zählerstand
            # gespeichert, falls er mindestens einmal übertragen wurde.
            log.output(2,'Stecker steckt, es liegt aber noch kein Zählerwert vor.')
            self.__counterAtPlugIn = self.__counter
            log.output(2,'Zählerstand bei Anstecken wird auf aktuellen Zählerstand gesetzt: '
                    +str(self.__counter))
            
        #Zustand des Steckers sichern
        self.__plug = s
