#!/usr/bin/env python3

"""
soc_helper
Copyright (C) 2023-2024  M. Williges (spam at zut punkt de)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# System-Importe
import paho.mqtt.client as mqtt
import paho.mqtt.publish as publish
import json
import sys
import time
import importlib
from os.path import exists
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

# Importe von lokalen Dateien
import log
import configuration	#Konfiguration des soc_helper.py
import ladevorgangclass #Daten und Methoden des aktuellen Ladevorgangs

# Hilfsroutinen für Hauptprogramm    
def openLocalLogfile(path):
    # Versuche, Datei unter 'path' ergänzend zu öffnen. 
    # Rückgabe ist ein Tupel (Status,file). Ein Rückgabewert von >0 bedeutet, daß das Öffnen nicht funktioniert hat.
    global savefile
    try:
        if not exists(path):
            log.output(1, path+' existiert nicht. Lege neu an.')
            savefile = open(path,'a')
            savefile.write('Datum, Kilometerstand, Energiemenge, Start-SOC (Auto), End-SOC (Wallbox)\n')
            savefile.flush()
        else:
            log.output(1, 'Öffne existierende Datei '+path)
            savefile = open(path,'a')
        return 0
    except:
        log.output(0,'Konnte Ausgabedatei '+path+' nicht anlegen oder öffnen. Warte auf Änderung von configuration.py')
        return 1

def closeLocalLogfile():
    global savefile
    savefile.close()

#
# Callback-Funktionen
#

# Konfigurationsdatei geändert, Logdatei und Verbindungen schließen, Konfiguration neu einlesen.
def cb_cfChanged(event):
    global client
    log.output(1,"configuration.py geändert, starte Neueinlesen.")
    log.output(2,"Schließe lokale Log-Datei "+configuration.CHARGELOG_PATH)
    closeLocalLogfile()
    log.output(2,"Schließe MQTT-Verbindung zu "+configuration.OPENWB_IP)
    client.disconnect()
    log.output(2,"Lade configuration.py neu ein...")
    importlib.invalidate_caches()
    importlib.reload(configuration)
    ladevorgangclass.reloadConfig()
    log.reloadConfig()
    # (ggf geänderte) Logdatei wird in Hauptschleife geöffnet
    # (ggf geädnerte) MQTT-Verbindung wird in Hauptschleife hergestellt

# Erfolgreicher Verbindung zur OpenWB (CONNACK)
def on_connect(client, userdata, flags, rc):
    log.output(1,'Verbindung hergestellt zu '+configuration.OPENWB_IP+' mit Resultat '+str(rc))
    # Subscribing in on_connect() means that if we lose the connection and
    # reconnect then subscriptions will be renewed.
    client.subscribe('#')

# Energiezählerübertragung. In der Regel so hochfrequent, daß dem Empfang von Kilometerstand / SOC immer
# ein aktueller Zählerstand zugeordnet werden kann.
def cb_energycounter(client, userdata, msg):
    ladevorgang.setCounter(float(msg.payload))

# Steckerzustandes des Ladekabels.
def cb_plug(client, userdata, msg):
    if (b'true' in msg.payload) or (b'1' in msg.payload):
        ladevorgang.setPlug(1)
    else:
        ladevorgang.setPlug(0)

# Berechneter SOC aus der Wallbox
def cb_openwbsoc(client, userdata, msg):
    ladevorgang.setOpenwbsoc(msg.payload)
        
# Callback-Funktion für WiCAN-Statusbotschaften. Wenn ein status:online vom wican empfangen wird, wird
# unmittelbar danach eine SOC- und eine ODO-Abfrage geschickt.
def cb_status(client, userdata, msg):
    log.output(2,'Status: '+msg.topic+" "+str(msg.payload))
    # Statusbotschaft empfangen. Prüfen, ob online
    try:
        s=json.loads(msg.payload)
        if s['status'] == 'online':
            # Wenn das Fahrzeug online geht, kann der SOC und der Gesamtkm-Stand abgefragt werden
            log.output(2,'Fahrzeug ist online. Sende SOC- und DST-Anforderung')
            client.publish(configuration.TOPIC_PREFIX+'/can/tx', configuration.SOC_REQUEST)
            client.publish(configuration.TOPIC_PREFIX+'/can/tx', configuration.DST_REQUEST)
        else:
            log.output(2,'Fahrzeugstatus ist nicht <<online>>')
    except:
        log.output(0,'Statusnachricht vom WiCAN enthält kein <status>')

# Callback-Funktion für eine CAN-Rx-Botschaft. Prüft, ob es eine SOC- oder ODO-haltige Botschaft ist
# und sendet im Fall eines SOC den Wert zur OpenWB
def cb_rx(client, userdata, msg):
    # CAN-Botschaft empfangen. Prüfen, ob SOC und dann auswerten und and openWB pushen
    log.output(2,'Empfangene CAN-Botschaft: '+str(msg.payload))
    try:
        d = json.loads(msg.payload)     # String der Nutzlast als dict in d
        f = d['frame']                  # Datenframe in f ergibt eine Liste mit einem Element
        dd = f[0]                       # Das nullte und einzige Element der Liste ist ein dict und wird in dd kopiert
        ddd=dd['data']                  # Das Element data aus dem frame-dict enthält die Nutzbytes vom CAN als list
        if dd['id'] == configuration.VALID_SOC_ID:
            if ddd[1:1+len(configuration.VALID_SOC_ANSWER)] == configuration.VALID_SOC_ANSWER:
                soc=configuration.getSoc(ddd)
                log.output(1,'Fahrzeug-SOC ist '+str(soc))
                ladevorgang.setSoc(soc)
                log.output(2,"SOC-Wert von "+str(soc)+' an '+configuration.OPENWB_SETSOC_TOPIC+' schicken.')
                try:
                    client.publish(configuration.OPENWB_SETSOC_TOPIC, soc)     #SOC-Wert an die OpenWB schicken
                except:
                    log.output(0,"Schreiben des SOC an die Wallbox ist fehlgeschlagen.")
            else:
                log.output(1,'Botschaft ist keine gültige Antwort auf SOC-Anfrage')
        elif dd['id'] == configuration.VALID_DST_ID:
            if ddd[1:1+len(configuration.VALID_DST_ANSWER)] == configuration.VALID_DST_ANSWER:
                odo=configuration.getOdo(ddd)
                ladevorgang.setOdo(odo)
                log.output(1,'Fahrzeug-Kilometerstand ist '+str(odo))
            else:
                log.output(1,'Kilometerstand-Botschaft ist keine gültige Antwort auf Kilometerstand-Anfrage')
        else:
            log.output(1,'Empfangene Botschaft trägt keinen bekannten Identifier. Wird ignoriert.')
    except:
        log.output(0,'Fehler beim Parsen einer CAN-Botschaft vom WiCAN. Wird ignoriert')

#
# Hauptprogramm
#


# Globale Klasse für Ladedaten erstellen
ladevorgang = ladevorgangclass.ladevorgangclass()
# Überwachung der Konfigurationsdatei einrichten, damit sie bei Änderungen neu eingelesen wird
patterns = ["configuration.py"]
ignore_patterns = None
ignore_directories = True
case_sensitive = True
cfHandler = PatternMatchingEventHandler(patterns, ignore_patterns, ignore_directories, case_sensitive)
cfHandler.on_modified = cb_cfChanged
cfWatchdog = Observer()
cfWatchdog.schedule(cfHandler, ".", recursive = False)
cfWatchdog.start()
# MQTT-Client einrichten und mit dem Broker verbinden
client = mqtt.Client()
client.on_pre_connect = None
client.on_connect = on_connect
# MQTT-Topic-Abos einrichten und mit Callback-Funktionen verbinden
client.message_callback_add(configuration.OPENWB_CHARGECOUNTER_TOPIC, cb_energycounter)
client.message_callback_add(configuration.OPENWB_GETPLUGSTATE_TOPIC, cb_plug)
client.message_callback_add(configuration.OPENWB_GETSOC_TOPIC, cb_openwbsoc)
client.message_callback_add(configuration.TOPIC_PREFIX+'/status/#', cb_status)
client.message_callback_add(configuration.TOPIC_PREFIX+'/can/rx/#', cb_rx)

log.output(0,'Starte Version 2024-01-14')

while 0<1:
    # Die Schleife ist nötig, da beim Neueinlesen von configuration.py eine bestehende MQTT-Verbindung geschlossen wird
    # (Schließlich könnte eine andere Zieladresse konfiguriert worden sein.). Nach Beenden der Verbindung endet
    # client.loop_forever()

    # Datei zum Abspeichern der Ladedaten öffnen (könnte sich durch Konfigurationsänderung geändert haben)
    # Momentan terminiert das Programm in diesem Fall.
    # Versuche, die lokale Logdatei zu öffnen und verbleibe in einer Schleife, bis das klappt
    # (die Überwachung von configuration.py sollte bei Dateiänderungen ein Neueinlesen verursachen)
    rv = openLocalLogfile(configuration.CHARGELOG_PATH)
    if rv > 0:
        #Öffnen fehlgeschlagen. Endlosschleife, die durch Änderung von configuration.py unterbrochen wird:
        while 0 < 1:
            time.sleep(1)
    # Öffnen erfolgreich, Filehandle-Übergabe an Klasse
    ladevorgang.setLogfilehandle(savefile)

    # Verbindung zum MQTT-Broker in der Wallbox herstellen
    # Verbesserungsidee: Durch Änderung der Konfiguration ist es möglich, daß eine ungültige IP-Adresse anliegt.
    #		Dann sollte hier eine Logmeldung ausgegeben werden und gewartet werden, bis configuration sich erneut 
    #		verändert hat.
    log.output(2,"Verbindungsversuch mit MQTT-Broker der Wallbox unter Adresse "+configuration.OPENWB_IP)
    try:
        client.connect(configuration.OPENWB_IP, 1883, 60)
    except:
        log.output(0,"FATAL: Fehler beim Verbinden zum Broker auf "+configuration.OPENWB_IP+'. Programmende.')
        quit()
    client.loop_forever()
    time.sleep(1)	# Nach Ende der Verbindung zum MQTT-Broker der Wallbox kurz warten vor neuer Kontaktaufnahme, sonst Fehler
